/* Copyright 2006 - Morten Proschowsky */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "pz.h"

static PzModule *module;
static TWidget *wid;
extern ttk_screeninfo *ttk_screen;

#define SNAKE_MAX_LENGTH 200
#define SNAKE_MAX_ROTATE 6 
#define TIMER_VALUE 50
static int snake360_X[SNAKE_MAX_LENGTH];
static int snake360_Y[SNAKE_MAX_LENGTH];
static int snake360_rb_head = 0;
static int snake360_rb_tail = 0;
static int snake360_grow = 0;
static int snake360_direction = 0;
static int snake360_inputbuffer = 0;
static int snake360_playpause = 0;
static int snake360_dead = 0;
static double snake360_head_x = 0;
static double snake360_head_y = 0;
static int snake360_speed = 3;
static int snake360_food_x;
static int snake360_food_y;
static int snake360_points;


void snake360_rb_init() {
	snake360_rb_head = snake360_rb_tail = 0;
	int i;
	for (i=0; i<SNAKE_MAX_LENGTH; i++)
		snake360_X[i] = snake360_Y[i] = -1;
}

int snake360_rb_length() {
	int length = snake360_rb_head - snake360_rb_tail;
       	return (length<0?SNAKE_MAX_LENGTH+length:length);
}

void snake360_rb_add(int x, int y) {
	snake360_rb_head++;
	if (snake360_rb_head == SNAKE_MAX_LENGTH) snake360_rb_head = 0;
	snake360_X[snake360_rb_head] = x;
	snake360_Y[snake360_rb_head] = y;
}

void snake360_rb_replace(int x, int y) {
	snake360_rb_add(x, y);
	snake360_rb_tail++;
	if (snake360_rb_tail == SNAKE_MAX_LENGTH) snake360_rb_tail = 0;
}

void snake360_rb_fetch(int i, int *x, int *y) {
	i = snake360_rb_head - i;
	if (i < 0) i = i + SNAKE_MAX_LENGTH;
	*x = snake360_X[i];
	*y = snake360_Y[i];
}

int snake360_collision_objects(int x1, int y1, int x2, int y2, int d) {
	double c = sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
	return(c<d);
}
			

int snake360_collision_of(int x, int y, int d, int i) {
	int snake_length = snake360_rb_length();
	int sx, sy;
	for(; i<snake_length; i++) {
		snake360_rb_fetch(i, &sx, &sy);
		double c = sqrt((sx-x)*(sx-x)+(sy-y)*(sy-y));
		if (c<d) return 1;
	}
	return 0;
}

int snake360_collision(int x, int y, int d) {return snake360_collision_of(x,y,d,0);}

int snake360_hitwall(int d) {
	return (wid->x+d > snake360_head_x || wid->y+d > snake360_head_y || wid->x+wid->w-d < snake360_head_x || wid->y+wid->h-d < snake360_head_y);
}	

void snake360_place_food() {
	do
	{
		snake360_food_x = wid->x+7+(rand()/(RAND_MAX/(wid->w-14)));
		snake360_food_y = wid->y+7+(rand()/(RAND_MAX/(wid->h-14)));
		printf("Testing food place (%i, %i)\n", snake360_food_x, snake360_food_y);
	}	
	while (snake360_collision(snake360_food_x, snake360_food_y, 9));
}
		
void snake360_draw (PzWidget *wid, ttk_surface srf) 
{
	// pause message
	char pausemsg[] = "Press play to start";
	
	// draw background
	ttk_fillrect(srf, wid->x, wid->y, wid->x+wid->w, wid->y+wid->h, ttk_makecol(200,200,240));
        ttk_rect(srf, wid->x, wid->y, wid->x+wid->w, wid->y+wid->h, ttk_makecol(100,100,100));
	ttk_rect(srf, wid->x+1, wid->y+1, wid->x+wid->w-1, wid->y+wid->h-1, ttk_makecol(50,50,50));       
	ttk_rect(srf, wid->x+2, wid->y+2, wid->x+wid->w-2, wid->y+wid->h-2, ttk_makecol(0,0,0));
	
	ttk_color snake_head, snake_body1, snake_body2, food;
        if (ttk_screen->bpp > 2) {
		snake_head = ttk_makecol(255,0,0);
		snake_body1 = ttk_makecol(255,0,0);
		snake_body2 = ttk_makecol(0,0,0);
		food = ttk_makecol(0,0,255);
	}
	else
		snake_head = snake_body1 = snake_body2 = food = ttk_makecol(0,0,0);
	
	int snake_length = snake360_rb_length();
	int x, y;
	snake360_rb_fetch(0, &x, &y);
	ttk_fillellipse(srf, x, y, 6, 6, snake_head);
	int i;
	for(i=1; i<snake_length; i++) {
		snake360_rb_fetch(i, &x, &y);
		ttk_fillellipse(srf, x, y, 5, 5, (i%2==0?snake_body1:snake_body2));
	}
	ttk_fillrect(srf, snake360_food_x-4, snake360_food_y-4,snake360_food_x+4, snake360_food_y+4, food);
	if (!snake360_playpause)
		ttk_text (srf, ttk_menufont, wid->x + (wid->w - ttk_text_width (ttk_menufont, pausemsg))/2, wid->y + wid->h - ttk_text_height (ttk_menufont) - 5, ttk_makecol (BLACK), pausemsg);
}
    

void snake360_updateposition(int rotation) {
	snake360_direction = (snake360_direction + rotation) % 96;
	double i = snake360_direction;
	i = 2*3.14159265*i/96;
	snake360_head_x += snake360_speed*cos(i);
	snake360_head_y += snake360_speed*sin(i);
	if (snake360_grow) {
		snake360_grow--;
		snake360_rb_add((int)snake360_head_x, (int)snake360_head_y);
	}
	else
	{ 
		snake360_rb_replace((int)snake360_head_x, (int)snake360_head_y);
	}	
}

void snake360_startgame() {
        int l = 10;
        int x = wid->x+wid->w/2 - l*snake360_speed/2;
        int y = wid->y+wid->h/2;
        snake360_rb_init();
        snake360_inputbuffer = 0;
        snake360_direction = 0;
        snake360_rb_add(x, y);
        snake360_head_x = x;
        snake360_head_y = y;
        snake360_grow = l;
        snake360_points = 0;
        int i;
        for (i=0;i<l;i++)
                snake360_updateposition(0);
        snake360_playpause = 0;
        snake360_dead = 0;
        snake360_place_food();
        printf("Starting game at (%i,%i)\n", x, y);
}

int snake360_event (PzEvent *ev) 
{
	int rv = TTK_EV_UNUSED;
    	switch (ev->type) {
		case PZ_EVENT_SCROLL:
			if (snake360_playpause) {
				if (snake360_inputbuffer * ev->arg < 0-SNAKE_MAX_ROTATE)
					snake360_inputbuffer = ev->arg;	
				else
					snake360_inputbuffer += ev->arg;
			}
			rv = TTK_EV_CLICK;
			break;
		case PZ_EVENT_BUTTON_UP:
			switch (ev->arg) {
				case PZ_BUTTON_MENU:
					pz_close_window (ev->wid->win);
					rv = TTK_EV_CLICK;
					break;
				case PZ_BUTTON_PLAY:
					snake360_playpause = (snake360_playpause?0:1);
					wid->dirty = 1;
					rv = TTK_EV_CLICK;
					break;
				case PZ_BUTTON_NEXT:
					snake360_speed++;
					break;
				case PZ_BUTTON_PREVIOUS:
					snake360_speed--;
					if (snake360_speed == 0) snake360_speed = 1;
					break;
			}
			break;
		case PZ_EVENT_TIMER:
			if (snake360_playpause) {
				int rotation = snake360_inputbuffer;
				if (rotation > SNAKE_MAX_ROTATE) rotation = SNAKE_MAX_ROTATE;
				if (rotation < -SNAKE_MAX_ROTATE) rotation = -SNAKE_MAX_ROTATE;
				snake360_inputbuffer -= rotation;
				
				snake360_updateposition(rotation);
				if (snake360_hitwall(8))
					snake360_dead = 1;
				
				if (snake360_collision_of((int)snake360_head_x,(int) snake360_head_y,9,18-snake360_speed)) snake360_dead =1;
			
				if (snake360_dead) {
					snake360_playpause = 0;
					char pmsg[30];
					sprintf(pmsg, "You got %i points.", snake360_points);
					pz_message_title("  GAME OVER", pmsg);
					snake360_startgame();
				}
				if (snake360_collision_objects((int)snake360_head_x,(int) snake360_head_y,snake360_food_x, snake360_food_y, 10)) {
					snake360_points++;
					snake360_grow = 4;
					snake360_place_food();
				}
				ev->wid->dirty = 1;
			}
			rv = TTK_EV_UNUSED;
			break;
    	}
    	return rv;
}

PzWindow *snake360_start()
{
    PzWindow *ret;
    ret = pz_new_window ("Snake360", PZ_WINDOW_NORMAL);
    wid = pz_add_widget (ret, snake360_draw, snake360_event);
    pz_widget_set_timer(wid, TIMER_VALUE);
    snake360_startgame();
    return pz_finish_window (ret);
}

void snake360_cleanup() 
{
    printf ("Bye-bye now...\n");
}

void init_snake360() 
{
    module = pz_register_module ("snake360", snake360_cleanup);
    pz_menu_add_action ("/Extra/Games/Snake360", snake360_start);
    printf ("Starting up...\n");
}

PZ_MOD_INIT (init_snake360)
